package inf101.v18.rogue101.shared;

import inf101.v18.grid.GridDirection;
import inf101.v18.grid.ILocation;
import inf101.v18.rogue101.enemies.Girl;
import inf101.v18.rogue101.game.IGame;
import inf101.v18.rogue101.objects.IActor;
import inf101.v18.rogue101.objects.IItem;
import inf101.v18.rogue101.states.Occupation;
import javafx.scene.media.Media;
import javafx.scene.media.MediaPlayer;

import java.io.File;
import java.util.List;

public class NPC {

    /**
     * Makes a NPC attack anything it can, or move closer to any enemy in its line of sight.
     *
     * @param game  An IGame object
     * @return      True if the NPC made a move. False otherwise
     */
    public static boolean tryAttack(IGame game) {
        List<ILocation> neighbours = game.getVisible();
        for (ILocation neighbour : neighbours) {
            if (game.getMap().hasActors(neighbour)) {
                ILocation current = game.getLocation();
                GridDirection dir = game.locationDirection(current, neighbour);
                IActor actor = game.getMap().getActors(neighbour).get(0); //We assume there is only one actor.
                if (current.gridDistanceTo(neighbour) <= 1 && current.canGo(dir) && game.getMap().has(current.go(dir), actor) && !actor.isDestroyed()) {
                    playSound("audio/Sword Swing-SoundBible.com-639083727.wav");
                    game.attack(dir, actor);
                    return true;
                } else if (game.canGo(dir)) {
                    game.move(dir);
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Performs a ranged attack on the closest actor (if any) observable by the current actor.
     *
     * @param game      An IGame object
     * @param range     The range of the equipped weapon
     * @return          True if an attack or a move towards an enemy was successful. False otherwise
     */
    public static boolean tryAttackRanged(IGame game, int range) {
        // Ranged attacks don't care about walls.
        List<ILocation> neighbours = game.getMap().getNeighbourhood(game.getLocation(), game.getActor().getVision());
        for (ILocation neighbour : neighbours) {
            if (game.getMap().hasActors(neighbour)) {
                ILocation current = game.getLocation();
                GridDirection dir = game.locationDirection(current, neighbour);
                IActor actor = game.getMap().getActors(neighbour).get(0); //We assume there is only one actor.
                if (current.gridDistanceTo(neighbour) <= range && !actor.isDestroyed()) {
                    if (game.getActor() instanceof Girl) {
                        Girl girl = (Girl)game.getActor();
                        if (girl.getOccupation() == Occupation.BOWSMAN) {
                            playSound("audio/Bow_Fire_Arrow-Stephan_Schutze-2133929391.wav");
                        } else if (girl.getOccupation() == Occupation.MAGE) {
                            playSound("audio/Large Fireball-SoundBible.com-301502490.wav");
                        }
                    }
                    game.rangedAttack(dir, actor);
                    return true;
                } else if (game.canGo(dir)) {
                    game.move(dir);
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Tries to find and reach an item of a specified class.
     *
     * @param game      An IGame object
     * @param element   The class of the objects we are looking for
     * @return          True if an appropriate item was found in the range of the current actor. False otherwise.
     */
    public static boolean trackItem(IGame game, Class<?> element) {
        List<ILocation> neighbours = game.getVisible();
        for (ILocation neighbour : neighbours) {
            boolean hasItem = false;
            for (IItem item : game.getMap().getAll(neighbour)) {
                if (element.isInstance(item)) {
                    hasItem = true;
                }
            }
            if (hasItem) {
                ILocation current = game.getLocation();
                GridDirection dir = game.locationDirection(current, neighbour);
                if (game.canGo(dir)) {
                    game.move(dir);
                    return true;
                }
            }
        }
        return false;
    }

    public static boolean flee(IGame game) {
        List<ILocation> neighbours = game.getVisible();
        for (ILocation neighbour : neighbours) {
            if (game.getMap().hasActors(neighbour)) {
                ILocation current = game.getLocation();
                GridDirection dir = reverseDir(game.locationDirection(current, neighbour));
                if (game.canGo(dir)) {
                    game.move(dir);
                    return true;
                }
            }
        }
        return false;
    }

    private static GridDirection reverseDir(GridDirection dir) {
        if (dir == GridDirection.SOUTH)  {
            return GridDirection.NORTH;
        } else if (dir == GridDirection.NORTH) {
           return GridDirection.SOUTH;
        } else if (dir == GridDirection.WEST) {
            return GridDirection.EAST;
        } else {
            return GridDirection.WEST;
        }
    }

    public static void playSound(String filename) {
        Media sound = new Media(new File(filename).toURI().toString());
        MediaPlayer mediaPlayer = new MediaPlayer(sound);
        mediaPlayer.play();
    }
}
